
/* EQUATION SPECIFIC FVM routines
 * MHD (Magneto-HydroDynamics equations)
 */

///::title:: MHD (Magneto-HydroDynamics equations)
///::consts:: NO_POWELL=0 GAMMA=(5.0/3.0)
///::vars:: POWELL=1

#include "equation.h"

void calc_point_flux(ConservedVars &F, const AllVariables &V, int d)
{
  real tP = V.XU.p + 0.5*V.U.B.sqr();

  F.rho = V.U.m[d];
  F.m = V.U.m * V.XU.u[d] - V.U.B * V.U.B[d];
  F.m[d] += tP;
  F.E = (V.U.E + tP) * V.XU.u[d] - (V.XU.u*V.U.B) * V.U.B[d];
  F.B = V.XU.u[d] * V.U.B - V.XU.u * V.U.B[d];
}

void apply_constraints (ConservedVars &CV, XtraVars &XV)
{
  // Constraints for ConservedVars
  if (CV.rho < 0.0) CV.rho = 0.0;
  if (CV.E   < 0.0) CV.E   = 0.0;
  // Constraints for XtraVars
  if (XV.p   < 0.0) XV.p   = 0.0;
}

real calc_time_step()
{
  vector maxeig;

  maxeig.zero();
  for (int i = 0; i < NXl; i++) {
    for (int j = 0; j < NYl; j++) {
      for (int k = 0; k < NZl; k++) {
        for (int d = 0; d < NDIR; d++) {
          if (!HAVE_DIR(d)) continue;
          maxeig[d] = max( maxeig[d], fabs(Xtra[i][j][k].u[d]) + Xtra[i][j][k].cf[d] );
        }
      }
    }
  }
  return CFL/(maxeig[X]/dx[X] + maxeig[Y]/dx[Y] + maxeig[Z]/dx[Z]);
}

void calc_xtra(XtraVars &Xijk, const ConservedVars &Uijk, int i, int j, int k)
{
  real ie = Uijk.E - 0.5*Uijk.m.sqr()/Uijk.rho - 0.5*Uijk.B.sqr();
  Xijk.u = Uijk.m / Uijk.rho;
  Xijk.p = (GAMMA-1)*ie;
  if (Xijk.p < 0) {
    std::cerr<<"Negative Pressure p="<<Xijk.p<<"("<<i<<' '<<j<<' '<<k<<") t="<<t<<"\n";
    throw 1;
  }
  Xijk.a = sqrt(GAMMA*Xijk.p/Uijk.rho);
  real cm2 = 0.5*(Xijk.a*Xijk.a + Uijk.B.sqr()/Uijk.rho);
  for (int d=0; d<NDIR; d++) {
    real x = cm2*cm2 - Xijk.a*Xijk.a*Uijk.B[d]*Uijk.B[d]/Uijk.rho;
    Xijk.cf[d] = sqrt(cm2 + ssqrt(x));
  }
}

void calc_all_vars(AllVariables &V, const PrimitiveVars &P)
{
  P.toConservedVars(V.U);
  V.XU.p = P.p;
  V.XU.u = P.u;
  if (P.p < 0) {
    std::cerr<<"Negative Pressure\n";
    throw 1;
  }
  V.XU.a = sqrt(GAMMA*P.p/P.rho);
  real cm2 = 0.5*(V.XU.a*V.XU.a + P.B.sqr()/P.rho);
  for (int d=0; d<NDIR; d++) {
    real x = cm2*cm2 - V.XU.a*V.XU.a*P.B[d]*P.B[d]/P.rho;
    V.XU.cf[d] = sqrt(cm2 + ssqrt(x));
  }
}

#if INDICATOR == 1
// rho + B^2
real indicator(int i, int j, int k, int d)
{
  ConservedVars &Uijk = U[i][j][k];
  return Uijk.rho + Uijk.B.sqr();
}
#endif
