
/* EQUATION SPECIFIC FVM routines
 * Shallow Water equations
 */

///::title:: Shallow Water Equations
///::vars:: GRAVITY=9.812 BED_OFFSET=0.0 CFL=0.4 EC_TOPOGRAPHY_QUAD=1
///::depends:: topography

#include "equation.h"

void calc_point_flux(ConservedVars &F, const AllVariables &V, int d)
{
  if (d != 2) {
    F.h       = V.U.h * V.XU.u[d];
    F.hu[d]   = V.U.h * sqr(V.XU.u[d]) + GRAVITY * sqr(V.U.h) / 2.0;
    F.hu[1-d] = V.U.h * V.XU.u[d] * V.XU.u[1-d];
  }
}

void apply_constraints (ConservedVars &CV, XtraVars &XV)
{
  // Constraints for ConservedVars
  if (CV.h < 0.0) CV.h = 0.0;
  // Constraints for XtraVars
  if (XV.E < 0.0) XV.E = 0.0;
}

real calc_time_step()
{
  vector maxeig;
  maxeig.zero();
  real sqrt_gh;
  
  for (int i = 0; i < NXl; i++)
    for (int j = 0; j < NYl; j++) {
      sqrt_gh = sqrt ( GRAVITY * U[i][j][0].h );
      for (int d = 0; d < NDIR; d++) {
          if (!HAVE_DIR(d)) continue;
          maxeig[d] = max( maxeig[d], fabs(Xtra[i][j][0].u[d]) + sqrt_gh );
      }
    }
  
  return CFL / ( maxeig[X]/dx[X] + maxeig[Y]/dx[Y] );
}

void calc_xtra(XtraVars &Xijk, const ConservedVars &Uijk, int i, int j, int k)
{
  Xijk.u = Uijk.hu / Uijk.h;
  Xijk.E = ( Uijk.h * (Xijk.u * Xijk.u) + GRAVITY * sqr(Uijk.h) ) / 2.0;
  Xijk.b = b_cell_center(i,j);
}

void calc_all_vars(AllVariables &V, const PrimitiveVars &P)
{
  P.toConservedVars(V.U);
  V.XU.u = P.u;
  V.XU.E = ( P.h * (P.u * P.u) + GRAVITY * sqr(P.h) ) / 2.0;
}

#if INDICATOR == 1
real indicator(int i, int j, int k, int d)
{
  ConservedVars &Uijk = U[i][j][k];
  return Uijk.h;
}
#endif

#if ENTROPIC_RECONSTRUCTION

void convert_to_EntropyVars (ConservedVars &CV, const XtraVars &XV) {
  CV.h += XV.b;
  CV.h *= GRAVITY;
  CV.h -= sqr(XV.u[X]) / 2.0;
  CV.h -= sqr(XV.u[Y]) / 2.0;
}

void convert_to_ConservedVars (ConservedVars &CV, const XtraVars &XV) {
  CV.h += sqr(XV.u[X]) / 2.0;
  CV.h += sqr(XV.u[Y]) / 2.0;
  CV.h /= GRAVITY;
  CV.h -= XV.b;
}

#endif
